# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Animals
"""

from webhelpers2.html import tags

from wuttaweb.forms.schema import WuttaDictEnum

from wuttafarm.db.model import AnimalType, AnimalAsset
from wuttafarm.web.views.assets import AssetTypeMasterView, AssetMasterView
from wuttafarm.web.forms.schema import AnimalTypeRef
from wuttafarm.web.forms.widgets import ImageWidget


class AnimalTypeView(AssetTypeMasterView):
    """
    Master view for Animal Types
    """

    model_class = AnimalType
    route_prefix = "animal_types"
    url_prefix = "/animal-types"

    farmos_refurl_path = "/admin/structure/taxonomy/manage/animal_type/overview"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "farmos_uuid",
        "drupal_id",
    ]

    has_rows = True
    row_model_class = AnimalAsset
    rows_viewable = True

    row_grid_columns = [
        "asset_name",
        "sex",
        "is_sterile",
        "birthdate",
        "archived",
    ]

    rows_sort_defaults = "asset_name"

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_farmos_url(self, animal_type):
        return self.app.get_farmos_url(f"/taxonomy/term/{animal_type.drupal_id}")

    def get_xref_buttons(self, animal_type):
        buttons = super().get_xref_buttons(animal_type)

        if animal_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_animal_types.view", uuid=animal_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons

    def get_row_grid_data(self, animal_type):
        model = self.app.model
        session = self.Session()
        return (
            session.query(model.AnimalAsset)
            .join(model.Asset)
            .filter(model.AnimalAsset.animal_type == animal_type)
        )

    def configure_row_grid(self, grid):
        g = grid
        super().configure_row_grid(g)
        model = self.app.model
        enum = self.app.enum

        # asset_name
        g.set_link("asset_name")
        g.set_sorter("asset_name", model.Asset.asset_name)
        g.set_filter("asset_name", model.Asset.asset_name)

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", model.Asset.archived)
        g.set_filter("archived", model.Asset.archived)

    def get_row_action_url_view(self, animal, i):
        return self.request.route_url("animal_assets.view", uuid=animal.uuid)


class AnimalAssetView(AssetMasterView):
    """
    Master view for Animal Assets
    """

    model_class = AnimalAsset
    route_prefix = "animal_assets"
    url_prefix = "/assets/animal"

    farmos_refurl_path = "/assets/animal"

    labels = {
        "animal_type": "Species/Breed",
        "is_sterile": "Sterile",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "asset_name",
        "produces_eggs",
        "animal_type",
        "birthdate",
        "is_sterile",
        "sex",
        "archived",
    ]

    form_fields = [
        "asset_name",
        "animal_type",
        "birthdate",
        "produces_eggs",
        "sex",
        "is_sterile",
        "notes",
        "asset_type",
        "archived",
        "farmos_uuid",
        "drupal_id",
        "thumbnail_url",
        "image_url",
        "thumbnail",
        "image",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        enum = self.app.enum

        # animal_type
        g.set_joiner("animal_type", lambda q: q.join(model.AnimalType))
        g.set_sorter("animal_type", model.AnimalType.name)
        g.set_filter("animal_type", model.AnimalType.name)
        if self.farmos_style_grid_links:
            g.set_renderer("animal_type", self.render_animal_type_for_grid)
        else:
            g.set_link("animal_type")

        # birthdate
        g.set_renderer("birthdate", "date")

        # sex
        g.set_enum("sex", enum.ANIMAL_SEX)

    def render_animal_type_for_grid(self, animal, field, value):
        url = self.request.route_url("animal_types.view", uuid=animal.animal_type_uuid)
        return tags.link_to(value, url)

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        animal = f.model_instance

        # animal_type
        f.set_node("animal_type", AnimalTypeRef(self.request))

        # sex
        if self.viewing and animal.sex is None:
            pass  # TODO: dict enum widget does not handle null values well
        else:
            f.set_node("sex", WuttaDictEnum(self.request, enum.ANIMAL_SEX))
            f.set_required("sex", False)


def defaults(config, **kwargs):
    base = globals()

    AnimalTypeView = kwargs.get("AnimalTypeView", base["AnimalTypeView"])
    AnimalTypeView.defaults(config)

    AnimalAssetView = kwargs.get("AnimalAssetView", base["AnimalAssetView"])
    AnimalAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Farm Plants
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos.master import TaxonomyMasterView
from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import UsersType, StructureType, FarmOSPlantTypes
from wuttafarm.web.forms.widgets import ImageWidget


class PlantTypeView(TaxonomyMasterView):
    """
    Master view for Plant Types in farmOS.
    """

    model_name = "farmos_plant_type"
    model_title = "farmOS Plant Type"
    model_title_plural = "farmOS Plant Types"

    route_prefix = "farmos_plant_types"
    url_prefix = "/farmOS/plant-types"

    farmos_taxonomy_type = "plant_type"
    farmos_refurl_path = "/admin/structure/taxonomy/manage/plant_type/overview"

    def get_xref_buttons(self, plant_type):
        buttons = super().get_xref_buttons(plant_type)
        model = self.app.model
        session = self.Session()

        if wf_plant_type := (
            session.query(model.PlantType)
            .filter(model.PlantType.farmos_uuid == plant_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "plant_types.view", uuid=wf_plant_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class PlantAssetView(FarmOSMasterView):
    """
    Master view for farmOS Plant Assets
    """

    model_name = "farmos_plant_assets"
    model_title = "farmOS Plant Asset"
    model_title_plural = "farmOS Plant Assets"

    route_prefix = "farmos_plant_assets"
    url_prefix = "/farmOS/assets/plant"

    farmos_refurl_path = "/assets/plant"

    grid_columns = [
        "name",
        "archived",
    ]

    sort_defaults = "name"

    form_fields = [
        "name",
        "plant_types",
        "archived",
        "owners",
        "location",
        "notes",
        "raw_image_url",
        "large_image_url",
        "thumbnail_image_url",
        "image",
    ]

    def get_grid_data(self, columns=None, session=None):
        result = self.farmos_client.asset.get("plant")
        return [self.normalize_plant(a) for a in result["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")
        g.set_searchable("name")

        # archived
        g.set_renderer("archived", "boolean")

    def get_instance(self):

        plant = self.farmos_client.resource.get_id(
            "asset", "plant", self.request.matchdict["uuid"]
        )
        self.raw_json = plant

        # instance data
        data = self.normalize_plant(plant["data"])

        if relationships := plant["data"].get("relationships"):

            # add plant types
            if plant_type := relationships.get("plant_type"):
                if plant_type["data"]:
                    data["plant_types"] = []
                    for plant_type in plant_type["data"]:
                        plant_type = self.farmos_client.resource.get_id(
                            "taxonomy_term", "plant_type", plant_type["id"]
                        )
                        data["plant_types"].append(
                            {
                                "uuid": plant_type["data"]["id"],
                                "name": plant_type["data"]["attributes"]["name"],
                            }
                        )

            # add location
            if location := relationships.get("location"):
                if location["data"]:
                    location = self.farmos_client.resource.get_id(
                        "asset", "structure", location["data"][0]["id"]
                    )
                    data["location"] = {
                        "uuid": location["data"]["id"],
                        "name": location["data"]["attributes"]["name"],
                    }

            # add owners
            if owner := relationships.get("owner"):
                data["owners"] = []
                for owner_data in owner["data"]:
                    owner = self.farmos_client.resource.get_id(
                        "user", "user", owner_data["id"]
                    )
                    data["owners"].append(
                        {
                            "uuid": owner["data"]["id"],
                            "display_name": owner["data"]["attributes"]["display_name"],
                        }
                    )

            # add image urls
            if image := relationships.get("image"):
                if image["data"]:
                    image = self.farmos_client.resource.get_id(
                        "file", "file", image["data"][0]["id"]
                    )
                    data["raw_image_url"] = self.app.get_farmos_url(
                        image["data"]["attributes"]["uri"]["url"]
                    )
                    # nb. other styles available:  medium, wide
                    data["large_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["large"]
                    data["thumbnail_image_url"] = image["data"]["attributes"][
                        "image_style_uri"
                    ]["thumbnail"]

        return data

    def get_instance_title(self, plant):
        return plant["name"]

    def normalize_plant(self, plant):

        if notes := plant["attributes"]["notes"]:
            notes = notes["value"]

        if self.farmos_4x:
            archived = plant["attributes"]["archived"]
        else:
            archived = plant["attributes"]["status"] == "archived"

        return {
            "uuid": plant["id"],
            "drupal_id": plant["attributes"]["drupal_internal__id"],
            "name": plant["attributes"]["name"],
            "location": colander.null,  # TODO
            "archived": archived,
            "notes": notes or colander.null,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        plant = f.model_instance

        # plant_types
        f.set_node("plant_types", FarmOSPlantTypes(self.request))

        # location
        f.set_node("location", StructureType(self.request))

        # owners
        f.set_node("owners", UsersType(self.request))

        # notes
        f.set_widget("notes", "notes")

        # archived
        f.set_node("archived", colander.Boolean())

        # image
        if url := plant.get("large_image_url"):
            f.set_widget("image", ImageWidget("plant image"))
            f.set_default("image", url)

    def get_xref_buttons(self, plant):
        model = self.app.model
        session = self.Session()

        buttons = [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/asset/{plant['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]

        if wf_plant := (
            session.query(model.Asset)
            .filter(model.Asset.farmos_uuid == plant["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url("plant_assets.view", uuid=wf_plant.uuid),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    PlantTypeView = kwargs.get("PlantTypeView", base["PlantTypeView"])
    PlantTypeView.defaults(config)

    PlantAssetView = kwargs.get("PlantAssetView", base["PlantAssetView"])
    PlantAssetView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views to expose farmOS data directly
"""

from .master import FarmOSMasterView


def includeme(config):
    config.include("wuttafarm.web.views.farmos.users")
    config.include("wuttafarm.web.views.farmos.quantities")
    config.include("wuttafarm.web.views.farmos.asset_types")
    config.include("wuttafarm.web.views.farmos.units")
    config.include("wuttafarm.web.views.farmos.land_types")
    config.include("wuttafarm.web.views.farmos.land_assets")
    config.include("wuttafarm.web.views.farmos.structure_types")
    config.include("wuttafarm.web.views.farmos.structures")
    config.include("wuttafarm.web.views.farmos.animal_types")
    config.include("wuttafarm.web.views.farmos.animals")
    config.include("wuttafarm.web.views.farmos.groups")
    config.include("wuttafarm.web.views.farmos.plants")
    config.include("wuttafarm.web.views.farmos.log_types")
    config.include("wuttafarm.web.views.farmos.logs_activity")
    config.include("wuttafarm.web.views.farmos.logs_harvest")
    config.include("wuttafarm.web.views.farmos.logs_medical")
    config.include("wuttafarm.web.views.farmos.logs_observation")



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
View for farmOS Quantity Types
"""

import datetime

import colander

from wuttaweb.forms.schema import WuttaDateTime
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import FarmOSUnitRef


class QuantityTypeView(FarmOSMasterView):
    """
    View for farmOS Quantity Types
    """

    model_name = "farmos_quantity_type"
    model_title = "farmOS Quantity Type"
    model_title_plural = "farmOS Quantity Types"

    route_prefix = "farmos_quantity_types"
    url_prefix = "/farmOS/quantity-types"

    grid_columns = [
        "label",
        "description",
    ]

    sort_defaults = "label"

    form_fields = [
        "label",
        "description",
    ]

    def get_grid_data(self, columns=None, session=None):
        result = self.farmos_client.resource.get("quantity_type")
        return [self.normalize_quantity_type(t) for t in result["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # label
        g.set_link("label")
        g.set_searchable("label")

        # description
        g.set_searchable("description")

    def get_instance(self):
        result = self.farmos_client.resource.get_id(
            "quantity_type", "quantity_type", self.request.matchdict["uuid"]
        )
        self.raw_json = result
        return self.normalize_quantity_type(result["data"])

    def get_instance_title(self, quantity_type):
        return quantity_type["label"]

    def normalize_quantity_type(self, quantity_type):
        return {
            "uuid": quantity_type["id"],
            "drupal_id": quantity_type["attributes"]["drupal_internal__id"],
            "label": quantity_type["attributes"]["label"],
            "description": quantity_type["attributes"]["description"],
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # description
        f.set_widget("description", "notes")

    def get_xref_buttons(self, quantity_type):
        model = self.app.model
        session = self.Session()
        buttons = []

        if wf_quantity_type := (
            session.query(model.QuantityType)
            .filter(model.QuantityType.farmos_uuid == quantity_type["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "quantity_types.view", uuid=wf_quantity_type.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class QuantityMasterView(FarmOSMasterView):
    """
    Base class for Quantity views
    """

    farmos_quantity_type = None

    grid_columns = [
        "measure",
        "value",
        "label",
        "changed",
    ]

    sort_defaults = ("changed", "desc")

    form_fields = [
        "measure",
        "value",
        "units",
        "label",
        "created",
        "changed",
    ]

    def get_grid_data(self, columns=None, session=None):
        result = self.farmos_client.resource.get("quantity", self.farmos_quantity_type)
        return [self.normalize_quantity(t) for t in result["data"]]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # value
        g.set_link("value")

        # changed
        g.set_renderer("changed", "datetime")

    def get_instance(self):
        quantity = self.farmos_client.resource.get_id(
            "quantity", self.farmos_quantity_type, self.request.matchdict["uuid"]
        )
        self.raw_json = quantity

        data = self.normalize_quantity(quantity["data"])

        if relationships := quantity["data"].get("relationships"):

            # add units
            if units := relationships.get("units"):
                if units["data"]:
                    unit = self.farmos_client.resource.get_id(
                        "taxonomy_term", "unit", units["data"]["id"]
                    )
                    data["units"] = {
                        "uuid": unit["data"]["id"],
                        "name": unit["data"]["attributes"]["name"],
                    }

        return data

    def get_instance_title(self, quantity):
        return quantity["value"]

    def normalize_quantity(self, quantity):

        if created := quantity["attributes"]["created"]:
            created = datetime.datetime.fromisoformat(created)
            created = self.app.localtime(created)

        if changed := quantity["attributes"]["changed"]:
            changed = datetime.datetime.fromisoformat(changed)
            changed = self.app.localtime(changed)

        return {
            "uuid": quantity["id"],
            "drupal_id": quantity["attributes"]["drupal_internal__id"],
            "measure": quantity["attributes"]["measure"],
            "value": quantity["attributes"]["value"],
            "label": quantity["attributes"]["label"] or colander.null,
            "created": created,
            "changed": changed,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)

        # created
        f.set_node("created", WuttaDateTime(self.request))
        f.set_widget("created", WuttaDateTimeWidget(self.request))

        # changed
        f.set_node("changed", WuttaDateTime(self.request))
        f.set_widget("changed", WuttaDateTimeWidget(self.request))

        # units
        f.set_node("units", FarmOSUnitRef())


class StandardQuantityView(QuantityMasterView):
    """
    View for farmOS Standard Quantities
    """

    model_name = "farmos_standard_quantity"
    model_title = "farmOS Standard Quantity"
    model_title_plural = "farmOS Standard Quantities"

    route_prefix = "farmos_quantities_standard"
    url_prefix = "/farmOS/quantities/standard"

    farmos_quantity_type = "standard"

    def get_xref_buttons(self, standard_quantity):
        model = self.app.model
        session = self.Session()
        buttons = []

        if wf_standard_quantity := (
            session.query(model.StandardQuantity)
            .join(model.Quantity)
            .filter(model.Quantity.farmos_uuid == standard_quantity["uuid"])
            .first()
        ):
            buttons.append(
                self.make_button(
                    f"View {self.app.get_title()} record",
                    primary=True,
                    url=self.request.route_url(
                        "quantities_standard.view", uuid=wf_standard_quantity.uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    QuantityTypeView = kwargs.get("QuantityTypeView", base["QuantityTypeView"])
    QuantityTypeView.defaults(config)

    StandardQuantityView = kwargs.get(
        "StandardQuantityView", base["StandardQuantityView"]
    )
    StandardQuantityView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Units
"""

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import Measure, Unit


class MeasureView(WuttaFarmMasterView):
    """
    Master view for Measures
    """

    model_class = Measure
    route_prefix = "measures"
    url_prefix = "/measures"

    grid_columns = [
        "name",
        "drupal_id",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "drupal_id",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")


class UnitView(WuttaFarmMasterView):
    """
    Master view for Units
    """

    model_class = Unit
    route_prefix = "units"
    url_prefix = "/units"

    farmos_refurl_path = "/admin/structure/taxonomy/manage/unit/overview"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "farmos_uuid",
        "drupal_id",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_farmos_url(self, unit):
        return self.app.get_farmos_url(f"/taxonomy/term/{unit.drupal_id}")

    def get_xref_buttons(self, unit):
        buttons = super().get_xref_buttons(unit)

        if unit.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_units.view", uuid=unit.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    MeasureView = kwargs.get("MeasureView", base["MeasureView"])
    MeasureView.defaults(config)

    UnitView = kwargs.get("UnitView", base["UnitView"])
    UnitView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base views for Logs
"""

from collections import OrderedDict

import colander

from wuttaweb.forms.schema import WuttaDictEnum
from wuttaweb.db import Session
from wuttaweb.forms.widgets import WuttaDateTimeWidget

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import LogType, Log
from wuttafarm.web.forms.schema import LogAssetRefs


def get_log_type_enum(config):
    app = config.get_app()
    model = app.model
    session = Session()
    log_types = OrderedDict()
    query = session.query(model.LogType).order_by(model.LogType.name)
    for log_type in query:
        log_types[log_type.drupal_id] = log_type.name
    return log_types


class LogTypeView(WuttaFarmMasterView):
    """
    Master view for Log Types
    """

    model_class = LogType
    route_prefix = "log_types"
    url_prefix = "/log-types"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "farmos_uuid",
        "drupal_id",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_xref_buttons(self, log_type):
        buttons = super().get_xref_buttons(log_type)

        if log_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_log_types.view", uuid=log_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class LogView(WuttaFarmMasterView):
    """
    Master view for All Logs
    """

    model_class = Log
    route_prefix = "log"
    url_prefix = "/logs"

    farmos_refurl_path = "/logs"

    viewable = False
    creatable = False
    editable = False
    deletable = False
    model_is_versioned = False

    labels = {
        "message": "Log Name",
    }

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "message",
        "log_type",
        "assets",
        "location",
        "quantity",
        "groups",
        "is_group_assignment",
    ]

    sort_defaults = ("timestamp", "desc")

    filter_defaults = {
        "message": {"active": True, "verb": "contains"},
    }

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)

        # timestamp
        g.set_renderer("timestamp", "date")
        g.set_link("timestamp")

        # message
        g.set_link("message")

        # log_type
        g.set_enum("log_type", get_log_type_enum(self.config))

        # assets
        g.set_renderer("assets", self.render_assets_for_grid)

        # view action links to final log record
        def log_url(log, i):
            return self.request.route_url(f"logs_{log.log_type}.view", uuid=log.uuid)

        g.add_action("view", icon="eye", url=log_url)

    def render_assets_for_grid(self, log, field, value):
        assets = [str(a.asset) for a in log._assets]
        return ", ".join(assets)


class LogMasterView(WuttaFarmMasterView):
    """
    Base class for Asset master views
    """

    labels = {
        "message": "Log Name",
        "owners": "Owner",
    }

    grid_columns = [
        "status",
        "drupal_id",
        "timestamp",
        "message",
        "assets",
        # "location",
        "quantity",
        "is_group_assignment",
        "owners",
    ]

    sort_defaults = ("timestamp", "desc")

    filter_defaults = {
        "message": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "message",
        "timestamp",
        "assets",
        "location",
        "quantity",
        "notes",
        "status",
        "log_type",
        "owners",
        "is_group_assignment",
        "farmos_uuid",
        "drupal_id",
    ]

    def get_query(self, session=None):
        """ """
        model = self.app.model
        model_class = self.get_model_class()
        session = session or self.Session()
        return session.query(model_class).join(model.Log)

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model

        # status
        g.set_sorter("status", model.Log.status)
        g.set_filter("status", model.Log.status)

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", model.Log.drupal_id)
        g.set_filter("drupal_id", model.Log.drupal_id)

        # timestamp
        g.set_renderer("timestamp", "date")
        g.set_link("timestamp")
        g.set_sorter("timestamp", model.Log.timestamp)
        g.set_filter("timestamp", model.Log.timestamp)

        # message
        g.set_link("message")
        g.set_sorter("message", model.Log.message)
        g.set_filter("message", model.Log.message)

        # assets
        g.set_renderer("assets", self.render_assets_for_grid)

    def render_assets_for_grid(self, log, field, value):
        return ", ".join([a.asset.asset_name for a in log.log._assets])

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        enum = self.app.enum
        log = f.model_instance

        # timestamp
        # TODO: the widget should be automatic (assn proxy field)
        f.set_widget("timestamp", WuttaDateTimeWidget(self.request))
        if self.creating:
            f.set_default("timestamp", self.app.make_utc())

        # assets
        if self.creating or self.editing:
            f.remove("assets")  # TODO: need to support this
        else:
            f.set_node("assets", LogAssetRefs(self.request))
            f.set_default("assets", [a.asset_uuid for a in log.log._assets])

        # location
        if self.creating or self.editing:
            f.remove("location")  # TODO: need to support this

        # log_type
        if self.creating:
            f.remove("log_type")
        else:
            f.set_node(
                "log_type",
                WuttaDictEnum(self.request, get_log_type_enum(self.config)),
            )
            f.set_readonly("log_type")

        # quantity
        if self.creating or self.editing:
            f.remove("quantity")  # TODO: need to support this

        # notes
        f.set_widget("notes", "notes")

        # owners
        if self.creating or self.editing:
            f.remove("owners")  # TODO: need to support this

        # status
        f.set_node("status", WuttaDictEnum(self.request, enum.LOG_STATUS))

        # is_group_assignment
        f.set_node("is_group_assignment", colander.Boolean())

    def objectify(self, form):
        log = super().objectify(form)

        if self.creating:
            model_class = self.get_model_class()
            log.log_type = self.get_farmos_log_type()

        return log

    def get_farmos_url(self, log):
        return self.app.get_farmos_url(f"/log/{log.drupal_id}")

    def get_farmos_log_type(self):
        return self.model_class.__wutta_hint__["farmos_log_type"]

    def get_xref_buttons(self, log):
        buttons = super().get_xref_buttons(log)

        if log.farmos_uuid:
            log_type = self.get_farmos_log_type()
            route = f"farmos_logs_{log_type}.view"
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(route, uuid=log.farmos_uuid),
                    icon_left="eye",
                )
            )

        return buttons


def defaults(config, **kwargs):
    base = globals()

    LogTypeView = kwargs.get("LogTypeView", base["LogTypeView"])
    LogTypeView.defaults(config)

    LogView = kwargs.get("LogView", base["LogView"])
    LogView.defaults(config)


def includeme(config):
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Custom form widgets for WuttaFarm
"""

import json

import colander
from deform.widget import Widget, SelectWidget
from webhelpers2.html import HTML, tags

from wuttaweb.forms.widgets import WuttaCheckboxChoiceWidget
from wuttaweb.db import Session

from wuttafarm.web.util import render_quantity_objects


class ImageWidget(Widget):
    """
    Widget to display an image URL for a record.
    """

    def __init__(self, alt_text, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alt_text = alt_text

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            return tags.image(cstruct, self.alt_text, **kw)

        return super().serialize(field, cstruct, **kw)


class LogQuickWidget(Widget):
    """
    Widget to display an image URL for a record.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            items = []
            for quick in json.loads(cstruct):
                items.append(HTML.tag("li", c=quick))
            return HTML.tag("ul", c=items)

        return super().serialize(field, cstruct, **kw)


class FarmOSRefWidget(SelectWidget):
    """
    Generic widget to display "any reference field" - as a link to
    view the farmOS record it references.  Only used by the farmOS
    direct API views.
    """

    def __init__(self, request, route_prefix, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.route_prefix = route_prefix

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            try:
                obj = json.loads(cstruct)
            except json.JSONDecodeError:
                name = dict(self.values)[cstruct]
                obj = {"uuid": cstruct, "name": name}

            return tags.link_to(
                obj["name"],
                self.request.route_url(f"{self.route_prefix}.view", uuid=obj["uuid"]),
            )

        return super().serialize(field, cstruct, **kw)


class FarmOSRefsWidget(Widget):

    def __init__(self, request, route_prefix, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.route_prefix = route_prefix

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            links = []
            for obj in json.loads(cstruct):
                url = self.request.route_url(
                    f"{self.route_prefix}.view", uuid=obj["uuid"]
                )
                links.append(HTML.tag("li", c=tags.link_to(obj["name"], url)))

            return HTML.tag("ul", c=links)

        return super().serialize(field, cstruct, **kw)


class FarmOSAssetRefsWidget(Widget):
    """
    Widget to display a "Assets" field for an asset.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            assets = []
            for asset in json.loads(cstruct):
                url = self.request.route_url(
                    f"farmos_{asset['asset_type']}_assets.view", uuid=asset["uuid"]
                )
                assets.append(HTML.tag("li", c=tags.link_to(asset["name"], url)))

            return HTML.tag("ul", c=assets)

        return super().serialize(field, cstruct, **kw)


class FarmOSLocationRefsWidget(Widget):
    """
    Widget to display a "Locations" field for an asset.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            locations = []
            for location in json.loads(cstruct):
                asset_type = location["type"].split("--")[1]
                url = self.request.route_url(
                    f"farmos_{asset_type}_assets.view", uuid=location["uuid"]
                )
                locations.append(HTML.tag("li", c=tags.link_to(location["name"], url)))

            return HTML.tag("ul", c=locations)

        return super().serialize(field, cstruct, **kw)


class FarmOSQuantityRefsWidget(Widget):
    """
    Widget to display a "Quantities" field for a log.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            quantities = json.loads(cstruct)
            return render_quantity_objects(quantities)

        return super().serialize(field, cstruct, **kw)


class FarmOSUnitRefWidget(Widget):
    """
    Widget to display a "Units" field for a quantity.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            unit = json.loads(cstruct)
            return unit["name"]

        return super().serialize(field, cstruct, **kw)


class AnimalTypeWidget(Widget):
    """
    Widget to display an "animal type" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            animal_type = json.loads(cstruct)
            return tags.link_to(
                animal_type["name"],
                self.request.route_url(
                    "farmos_animal_types.view", uuid=animal_type["uuid"]
                ),
            )

        return super().serialize(field, cstruct, **kw)


class FarmOSPlantTypesWidget(Widget):
    """
    Widget to display a farmOS "plant types" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            links = []
            for plant_type in json.loads(cstruct):
                link = tags.link_to(
                    plant_type["name"],
                    self.request.route_url(
                        "farmos_plant_types.view", uuid=plant_type["uuid"]
                    ),
                )
                links.append(HTML.tag("li", c=link))
            return HTML.tag("ul", c=links)

        return super().serialize(field, cstruct, **kw)


class PlantTypeRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for Plant Types field (on a Plant Asset).
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            plant_types = []
            for uuid in json.loads(cstruct):
                plant_type = session.get(model.PlantType, uuid)
                plant_types.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(plant_type),
                            self.request.route_url(
                                "plant_types.view", uuid=plant_type.uuid
                            ),
                        ),
                    )
                )
            return HTML.tag("ul", c=plant_types)

        return super().serialize(field, cstruct, **kw)


class StructureWidget(Widget):
    """
    Widget to display a "structure" field.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            structure = json.loads(cstruct)
            return tags.link_to(
                structure["name"],
                self.request.route_url(
                    "farmos_structure_assets.view", uuid=structure["uuid"]
                ),
            )

        return super().serialize(field, cstruct, **kw)


class UsersWidget(Widget):
    """
    Widget to display the list of owners for an asset etc.
    """

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request

    def serialize(self, field, cstruct, **kw):
        """ """
        readonly = kw.get("readonly", self.readonly)
        if readonly:
            if cstruct in (colander.null, None):
                return HTML.tag("span")

            items = []
            for user in json.loads(cstruct):
                link = tags.link_to(
                    user["display_name"],
                    self.request.route_url("farmos_users.view", uuid=user["uuid"]),
                )
                items.append(HTML.tag("li", c=link))

            return HTML.tag("ul", c=items)

        return super().serialize(field, cstruct, **kw)


class AssetParentRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for Parents field which references assets.
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            parents = []
            for uuid in json.loads(cstruct):
                parent = session.get(model.Asset, uuid)
                parents.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(parent),
                            self.request.route_url(
                                f"{parent.asset_type}_assets.view", uuid=parent.uuid
                            ),
                        ),
                    )
                )
            return HTML.tag("ul", c=parents)

        return super().serialize(field, cstruct, **kw)


class LogAssetRefsWidget(WuttaCheckboxChoiceWidget):
    """
    Widget for Assets field (on a Log record)
    """

    def serialize(self, field, cstruct, **kw):
        """ """
        model = self.app.model
        session = Session()

        readonly = kw.get("readonly", self.readonly)
        if readonly:
            assets = []
            for uuid in json.loads(cstruct):
                asset = session.get(model.Asset, uuid)
                assets.append(
                    HTML.tag(
                        "li",
                        c=tags.link_to(
                            str(asset),
                            self.request.route_url(
                                f"{asset.asset_type}_assets.view", uuid=asset.uuid
                            ),
                        ),
                    )
                )
            return HTML.tag("ul", c=assets)

        return super().serialize(field, cstruct, **kw)



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for Asset master views
"""

import colander
from webhelpers2.html import tags

from wuttafarm.web.views.farmos import FarmOSMasterView
from wuttafarm.web.forms.schema import FarmOSRefs, FarmOSLocationRefs
from wuttafarm.web.forms.widgets import ImageWidget
from wuttafarm.web.grids import (
    ResourceData,
    StringFilter,
    IntegerFilter,
    BooleanFilter,
    SimpleSorter,
)


class AssetMasterView(FarmOSMasterView):
    """
    Base class for Asset master views
    """

    farmos_asset_type = None
    creatable = True
    editable = True
    deletable = True
    filterable = True
    sort_on_backend = True

    labels = {
        "name": "Asset Name",
        "asset_type_name": "Asset Type",
        "owners": "Owner",
        "locations": "Location",
        "thumbnail_url": "Thumbnail URL",
        "image_url": "Image URL",
    }

    grid_columns = [
        "thumbnail",
        "drupal_id",
        "name",
        "owners",
        "locations",
        "archived",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
        "archived": {"active": True, "verb": "is_false"},
    }

    def get_grid_data(self, **kwargs):
        return ResourceData(
            self.config,
            self.farmos_client,
            f"asset--{self.farmos_asset_type}",
            include=",".join(self.get_farmos_api_includes()),
            normalizer=self.normalize_asset,
        )

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # thumbnail
        g.set_renderer("thumbnail", self.render_grid_thumbnail)
        g.set_label("thumbnail", "", column_only=True)
        g.set_centered("thumbnail")

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", SimpleSorter("drupal_internal__id"))
        g.set_filter("drupal_id", IntegerFilter, path="drupal_internal__id")

        # name
        g.set_link("name")
        g.set_sorter("name", SimpleSorter("name"))
        g.set_filter("name", StringFilter)

        # owners
        g.set_renderer("owners", self.render_owners_for_grid)

        # locations
        g.set_renderer("locations", self.render_locations_for_grid)

        # archived
        g.set_renderer("archived", "boolean")
        g.set_sorter("archived", SimpleSorter("archived"))
        g.set_filter("archived", BooleanFilter)

    def render_grid_thumbnail(self, obj, field, value):
        if url := obj.get("thumbnail_url"):
            return tags.image(url, f"thumbnail for {self.get_model_title()}")
        return None

    def render_locations_for_grid(self, asset, field, value):
        locations = []
        for location in value:
            if self.farmos_style_grid_links:
                asset_type = location["type"].split("--")[1]
                route = f"farmos_{asset_type}_assets.view"
                url = self.request.route_url(route, uuid=location["uuid"])
                locations.append(tags.link_to(location["name"], url))
            else:
                locations.append(location["name"])
        return ", ".join(locations)

    def grid_row_class(self, asset, data, i):
        """ """
        if asset["archived"]:
            return "has-background-warning"
        return None

    def get_farmos_api_includes(self):
        return {"asset_type", "location", "owner", "image"}

    def get_instance(self):
        result = self.farmos_client.asset.get_id(
            self.farmos_asset_type,
            self.request.matchdict["uuid"],
            params={"include": ",".join(self.get_farmos_api_includes())},
        )
        self.raw_json = result
        included = {obj["id"]: obj for obj in result.get("included", [])}
        return self.normalize_asset(result["data"], included)

    def get_instance_title(self, asset):
        return asset["name"]

    def normalize_asset(self, asset, included):

        if notes := asset["attributes"]["notes"]:
            notes = notes["value"]

        if self.farmos_4x:
            archived = asset["attributes"]["archived"]
        else:
            archived = asset["attributes"]["status"] == "archived"

        asset_type_object = {}
        asset_type_name = None
        owner_objects = []
        owner_names = []
        location_objects = []
        location_names = []
        thumbnail_url = None
        image_url = None
        if relationships := asset.get("relationships"):

            if asset_type := relationships.get("asset_type"):
                if asset_type := included.get(asset_type["data"]["id"]):
                    asset_type_object = {
                        "uuid": asset_type["id"],
                        "name": asset_type["attributes"]["label"],
                    }
                    asset_type_name = asset_type_object["name"]

            if owners := relationships.get("owner"):
                for user in owners["data"]:
                    if user := included.get(user["id"]):
                        user = {
                            "uuid": user["id"],
                            "name": user["attributes"]["name"],
                        }
                        owner_objects.append(user)
                        owner_names.append(user["name"])

            if locations := relationships.get("location"):
                for location in locations["data"]:
                    if location := included.get(location["id"]):
                        location = {
                            "uuid": location["id"],
                            "type": location["type"],
                            "name": location["attributes"]["name"],
                        }
                        location_objects.append(location)
                        location_names.append(location["name"])

            if images := relationships.get("image"):
                for image in images["data"]:
                    if image := included.get(image["id"]):
                        thumbnail_url = image["attributes"]["image_style_uri"][
                            "thumbnail"
                        ]
                        image_url = image["attributes"]["image_style_uri"]["large"]

        return {
            "uuid": asset["id"],
            "drupal_id": asset["attributes"]["drupal_internal__id"],
            "name": asset["attributes"]["name"],
            "asset_type": asset_type_object,
            "asset_type_name": asset_type_name,
            "notes": notes or colander.null,
            "owners": owner_objects,
            "owner_names": owner_names,
            "locations": location_objects,
            "location_names": location_names,
            "archived": archived,
            "thumbnail_url": thumbnail_url or colander.null,
            "image_url": image_url or colander.null,
        }

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        asset = f.model_instance

        # asset_type_name
        if self.creating or self.editing:
            f.remove("asset_type_name")

        # locations
        if self.creating or self.editing:
            f.remove("locations")
        else:
            f.set_node("locations", FarmOSLocationRefs(self.request))

        # owners
        if self.creating or self.editing:
            f.remove("owners")  # TODO
        else:
            f.set_node("owners", FarmOSRefs(self.request, "farmos_users"))

        # notes
        f.set_widget("notes", "notes")
        f.set_required("notes", False)

        # archived
        f.set_node("archived", colander.Boolean())

        # thumbnail_url
        if self.creating or self.editing:
            f.remove("thumbnail_url")

        # image_url
        if self.creating or self.editing:
            f.remove("image_url")

        # thumbnail
        if self.creating or self.editing:
            f.remove("thumbnail")
        elif asset.get("thumbnail_url"):
            f.set_widget("thumbnail", ImageWidget("asset thumbnail"))
            f.set_default("thumbnail", asset["thumbnail_url"])

        # image
        if self.creating or self.editing:
            f.remove("image")
        elif asset.get("image_url"):
            f.set_widget("image", ImageWidget("asset image"))
            f.set_default("image", asset["image_url"])

    def persist(self, asset, session=None):
        payload = self.get_api_payload(asset)
        if self.editing:
            payload["id"] = asset["uuid"]

        result = self.farmos_client.asset.send(self.farmos_asset_type, payload)

        if self.creating:
            asset["uuid"] = result["data"]["id"]

    def get_api_payload(self, asset):

        attrs = {
            "name": asset["name"],
            "notes": {"value": asset["notes"] or None},
            "archived": asset["archived"],
        }

        if "is_location" in asset:
            attrs["is_location"] = asset["is_location"]

        if "is_fixed" in asset:
            attrs["is_fixed"] = asset["is_fixed"]

        return {"attributes": attrs}

    def delete_instance(self, asset):
        self.farmos_client.asset.delete(self.farmos_asset_type, asset["uuid"])

    def get_xref_buttons(self, asset):
        return [
            self.make_button(
                "View in farmOS",
                primary=True,
                url=self.app.get_farmos_url(f"/asset/{asset['drupal_id']}"),
                target="_blank",
                icon_left="external-link-alt",
            ),
        ]



# -*- coding: utf-8; -*-
################################################################################
#
#  WuttaFarm --Web app to integrate with and extend farmOS
#  Copyright © 2026 Lance Edgar
#
#  This file is part of WuttaFarm.
#
#  WuttaFarm is free software: you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  WuttaFarm is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
#  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along with
#  WuttaFarm.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Master view for Quantities
"""

from collections import OrderedDict

from wuttaweb.db import Session

from wuttafarm.web.views import WuttaFarmMasterView
from wuttafarm.db.model import QuantityType, Quantity, StandardQuantity
from wuttafarm.web.forms.schema import UnitRef


def get_quantity_type_enum(config):
    app = config.get_app()
    model = app.model
    session = Session()
    quantity_types = OrderedDict()
    query = session.query(model.QuantityType).order_by(model.QuantityType.name)
    for quantity_type in query:
        quantity_types[quantity_type.drupal_id] = quantity_type.name
    return quantity_types


class QuantityTypeView(WuttaFarmMasterView):
    """
    Master view for Quantity Types
    """

    model_class = QuantityType
    route_prefix = "quantity_types"
    url_prefix = "/quantity-types"

    grid_columns = [
        "name",
        "description",
    ]

    sort_defaults = "name"

    filter_defaults = {
        "name": {"active": True, "verb": "contains"},
    }

    form_fields = [
        "name",
        "description",
        "farmos_uuid",
        "drupal_id",
    ]

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)

        # name
        g.set_link("name")

    def get_xref_buttons(self, quantity_type):
        buttons = super().get_xref_buttons(quantity_type)

        if quantity_type.farmos_uuid:
            buttons.append(
                self.make_button(
                    "View farmOS record",
                    primary=True,
                    url=self.request.route_url(
                        "farmos_quantity_types.view", uuid=quantity_type.farmos_uuid
                    ),
                    icon_left="eye",
                )
            )

        return buttons


class QuantityMasterView(WuttaFarmMasterView):
    """
    Base class for Quantity master views
    """

    grid_columns = [
        "drupal_id",
        "as_text",
        "quantity_type",
        "measure",
        "value",
        "units",
        "label",
    ]

    sort_defaults = ("drupal_id", "desc")

    form_fields = [
        "quantity_type",
        "as_text",
        "measure",
        "value",
        "units",
        "label",
        "farmos_uuid",
        "drupal_id",
    ]

    def get_query(self, session=None):
        """ """
        model = self.app.model
        model_class = self.get_model_class()
        session = session or self.Session()
        query = session.query(model_class)
        if model_class is not model.Quantity:
            query = query.join(model.Quantity)
        query = query.join(model.Measure).join(model.Unit)
        return query

    def configure_grid(self, grid):
        g = grid
        super().configure_grid(g)
        model = self.app.model
        model_class = self.get_model_class()

        # drupal_id
        g.set_label("drupal_id", "ID", column_only=True)
        g.set_sorter("drupal_id", model.Quantity.drupal_id)

        # as_text
        g.set_renderer("as_text", self.render_as_text_for_grid)
        g.set_link("as_text")

        # quantity_type
        if model_class is not model.Quantity:
            g.remove("quantity_type")
        else:
            g.set_enum("quantity_type", get_quantity_type_enum(self.config))

        # measure
        g.set_sorter("measure", model.Measure.name)

        # value
        g.set_renderer("value", self.render_value_for_grid)

        # units
        g.set_sorter("units", model.Unit.name)

        # label
        g.set_sorter("label", model.Quantity.label)

        # view action links to final quantity record
        if model_class is model.Quantity:

            def quantity_url(quantity, i):
                return self.request.route_url(
                    f"quantities_{quantity.quantity_type_id}.view", uuid=quantity.uuid
                )

            g.add_action("view", icon="eye", url=quantity_url)

    def render_as_text_for_grid(self, quantity, field, value):
        return quantity.render_as_text(self.config)

    def render_value_for_grid(self, quantity, field, value):
        value = quantity.value_numerator / quantity.value_denominator
        return self.app.render_quantity(value)

    def get_instance_title(self, quantity):
        return quantity.render_as_text(self.config)

    def configure_form(self, form):
        f = form
        super().configure_form(f)
        quantity = form.model_instance

        # as_text
        if self.creating or self.editing:
            f.remove("as_text")
        else:
            f.set_default("as_text", quantity.render_as_text(self.config))

        # quantity_type
        if self.creating:
            f.remove("quantity_type")
        else:
            f.set_readonly("quantity_type")
            f.set_default("quantity_type", quantity.quantity_type.name)

        # measure
        if self.creating:
            f.remove("measure")
        else:
            f.set_readonly("measure")
            f.set_default("measure", quantity.measure.name)

        # value
        if self.creating:
            f.remove("value")
        else:
            value = quantity.value_numerator / quantity.value_denominator
            value = self.app.render_quantity(value)
            f.set_default(
                "value",
                f"{value} ({quantity.value_numerator} / {quantity.value_denominator})",
            )

        # units
        if self.creating:
            f.remove("units")
        else:
            f.set_readonly("units")
            f.set_node("units", UnitRef(self.request))
            # TODO: ugh
            f.set_default("units", quantity.quantity.units)

    def get_xref_buttons(self, quantity):
        buttons = super().get_xref_buttons(quantity)

        if quantity.farmos_uuid:
            url = self.request.route_url(
                f"farmos_quantities_{quantity.quantity_type_id}.view",
                uuid=quantity.farmos_uuid,
            )
            buttons.append(
                self.make_button(
                    "View farmOS record", primary=True, url=url, icon_left="eye"
                )
            )

        return buttons


class QuantityView(QuantityMasterView):
    """
    Master view for All Quantities
    """

    model_class = Quantity
    route_prefix = "quantities"
    url_prefix = "/quantities"

    viewable = False
    creatable = False
    editable = False
    deletable = False
    model_is_versioned = False


class StandardQuantityView(QuantityMasterView):
    """
    Master view for Standard Quantities
    """

    model_class = StandardQuantity
    route_prefix = "quantities_standard"
    url_prefix = "/quantities/standard"


def defaults(config, **kwargs):
    base = globals()

    QuantityTypeView = kwargs.get("QuantityTypeView", base["QuantityTypeView"])
    QuantityTypeView.defaults(config)

    QuantityView = kwargs.get("QuantityView", base["QuantityView"])
    QuantityView.defaults(config)

    StandardQuantityView = kwargs.get(
        "StandardQuantityView", base["StandardQuantityView"]
    )
    StandardQuantityView.defaults(config)


def includeme(config):
    defaults(config)
